@@ -168,16 +168,19 @@ class ActionClient(Waitable):
         # key: goal request sequence_number, value: Future for goal response
         self._pending_goal_requests = {}
         # key: goal request sequence_number, value: UUID
-        self._sequence_number_to_goal_id = {}
+        self._goal_sequence_number_to_goal_id = {}
         # key: cancel request sequence number, value: Future for cancel response
         self._pending_cancel_requests = {}
         # key: result request sequence number, value: Future for result response
         self._pending_result_requests = {}
+        # key: result request sequence_number, value: UUID
+        self._result_sequence_number_to_goal_id = {}
         # key: UUID in bytes, value: callback function
         self._feedback_callbacks = {}
 
         callback_group.add_entity(self)
         self._node.add_waitable(self)
+        self._logger = self._node.get_logger().get_child('action_client')
 
     def _generate_random_uuid(self):
         return UUID(uuid=list(uuid.uuid4().bytes))
@@ -208,14 +211,20 @@ class ActionClient(Waitable):
 
     def _remove_pending_goal_request(self, future):
         seq = self._remove_pending_request(future, self._pending_goal_requests)
-        if seq in self._sequence_number_to_goal_id:
-            del self._sequence_number_to_goal_id[seq]
+        if seq in self._goal_sequence_number_to_goal_id:
+            del self._goal_sequence_number_to_goal_id[seq]
 
     def _remove_pending_cancel_request(self, future):
         self._remove_pending_request(future, self._pending_cancel_requests)
 
     def _remove_pending_result_request(self, future):
-        self._remove_pending_request(future, self._pending_result_requests)
+        seq = self._remove_pending_request(future, self._pending_result_requests)
+        if seq in self._result_sequence_number_to_goal_id:
+            goal_uuid = bytes(self._result_sequence_number_to_goal_id[seq].uuid)
+            del self._result_sequence_number_to_goal_id[seq]
+            # remove feeback_callback if user is aware of result and it's been received
+            if goal_uuid in self._feedback_callbacks:
+                del self._feedback_callbacks[goal_uuid]
 
     # Start Waitable API
     def is_ready(self, wait_set):
@@ -277,10 +286,10 @@ class ActionClient(Waitable):
         """
         if 'goal' in taken_data:
             sequence_number, goal_response = taken_data['goal']
-            if sequence_number in self._sequence_number_to_goal_id:
+            if sequence_number in self._goal_sequence_number_to_goal_id:
                 goal_handle = ClientGoalHandle(
                     self,
-                    self._sequence_number_to_goal_id[sequence_number],
+                    self._goal_sequence_number_to_goal_id[sequence_number],
                     goal_response)
 
                 if goal_handle.accepted:
@@ -292,7 +301,7 @@ class ActionClient(Waitable):
 
                 self._pending_goal_requests[sequence_number].set_result(goal_handle)
             else:
-                self._node.get_logger().warning(
+                self._logger.warning(
                     'Ignoring unexpected goal response. There may be more than '
                     f"one action server for the action '{self._action_name}'"
                 )
@@ -302,7 +311,7 @@ class ActionClient(Waitable):
             if sequence_number in self._pending_cancel_requests:
                 self._pending_cancel_requests[sequence_number].set_result(cancel_response)
             else:
-                self._node.get_logger().warning(
+                self._logger.warning(
                     'Ignoring unexpected cancel response. There may be more than '
                     f"one action server for the action '{self._action_name}'"
                 )
@@ -312,7 +321,7 @@ class ActionClient(Waitable):
             if sequence_number in self._pending_result_requests:
                 self._pending_result_requests[sequence_number].set_result(result_response)
             else:
-                self._node.get_logger().warning(
+                self._logger.warning(
                     'Ignoring unexpected result response. There may be more than '
                     f"one action server for the action '{self._action_name}'"
                 )
@@ -351,6 +360,13 @@ class ActionClient(Waitable):
     def add_to_wait_set(self, wait_set):
         """Add entities to wait set."""
         self._client_handle.add_to_waitset(wait_set)
+
+    def __enter__(self):
+        return self._client_handle.__enter__()
+
+    def __exit__(self, t, v, tb):
+        self._client_handle.__exit__(t, v, tb)
+
     # End Waitable API
 
     def send_goal(self, goal, **kwargs):
@@ -433,7 +449,7 @@ class ActionClient(Waitable):
 
         future = Future()
         self._pending_goal_requests[sequence_number] = future
-        self._sequence_number_to_goal_id[sequence_number] = request.goal_id
+        self._goal_sequence_number_to_goal_id[sequence_number] = request.goal_id
         future.add_done_callback(self._remove_pending_goal_request)
         # Add future so executor is aware
         self.add_future(future)
@@ -538,6 +554,7 @@ class ActionClient(Waitable):
 
         future = Future()
         self._pending_result_requests[sequence_number] = future
+        self._result_sequence_number_to_goal_id[sequence_number] = result_request.goal_id
         future.add_done_callback(self._remove_pending_result_request)
         # Add future so executor is aware
         self.add_future(future)
@@ -576,13 +593,5 @@ class ActionClient(Waitable):
 
     def destroy(self):
         """Destroy the underlying action client handle."""
-        if self._client_handle is None:
-            return
-        with self._node.handle:
-            self._client_handle.destroy_when_not_in_use()
-            self._node.remove_waitable(self)
-        self._client_handle = None
-
-    def __del__(self):
-        """Destroy the underlying action client handle."""
-        self.destroy()
+        self._client_handle.destroy_when_not_in_use()
+        self._node.remove_waitable(self)
