@@ -70,8 +70,6 @@ class ServerGoalHandle:
         self._goal_info = goal_info
         self._goal_request = goal_request
         self._cancel_requested = False
-        self._result_future = Future()
-        action_server.add_future(self._result_future)
         self._lock = threading.Lock()
 
     def __eq__(self, other):
@@ -164,8 +162,6 @@ class ServerGoalHandle:
             self._goal_handle.destroy_when_not_in_use()
             self._goal_handle = None
 
-        self._action_server.remove_future(self._result_future)
-
 
 def default_handle_accepted_callback(goal_handle):
     """Execute the goal."""
@@ -260,8 +256,12 @@ class ActionServer(Waitable):
         # key: UUID in bytes, value: GoalHandle
         self._goal_handles = {}
 
+        # key: UUID in bytes, value: Future
+        self._result_futures = {}
+
         callback_group.add_entity(self)
         self._node.add_waitable(self)
+        self._logger = self._node.get_logger().get_child('action_server')
 
     async def _execute_goal_request(self, request_header_and_message):
         request_header, goal_request = request_header_and_message
@@ -269,7 +269,7 @@ class ActionServer(Waitable):
         goal_info = GoalInfo()
         goal_info.goal_id = goal_uuid
 
-        self._node.get_logger().debug('New goal request with ID: {0}'.format(goal_uuid.uuid))
+        self._logger.debug('New goal request with ID: {0}'.format(goal_uuid.uuid))
 
         # Check if goal ID is already being tracked by this action server
         with self._lock:
@@ -280,7 +280,7 @@ class ActionServer(Waitable):
             # Call user goal callback
             response = await await_or_execute(self._goal_callback, goal_request.goal)
             if not isinstance(response, GoalResponse):
-                self._node.get_logger().warning(
+                self._logger.warning(
                     'Goal request callback did not return a GoalResponse type. Rejecting goal.')
             else:
                 accepted = GoalResponse.ACCEPT == response
@@ -294,30 +294,39 @@ class ActionServer(Waitable):
                 with self._lock:
                     goal_handle = ServerGoalHandle(self, goal_info, goal_request.goal)
             except RuntimeError as e:
-                self._node.get_logger().error(
+                self._logger.error(
                     'Failed to accept new goal with ID {0}: {1}'.format(goal_uuid.uuid, e))
                 accepted = False
             else:
                 self._goal_handles[bytes(goal_uuid.uuid)] = goal_handle
+                self._result_futures[bytes(goal_uuid.uuid)] = Future()
+                self.add_future(self._result_futures[bytes(goal_uuid.uuid)])
 
         # Send response
         response_msg = self._action_type.Impl.SendGoalService.Response()
         response_msg.accepted = accepted
         response_msg.stamp = goal_info.stamp
-        self._handle.send_goal_response(request_header, response_msg)
 
-        if not accepted:
-            self._node.get_logger().debug('New goal rejected: {0}'.format(goal_uuid.uuid))
+        try:
+            # If the client goes away anytime before this, sending the goal response may fail.
+            # Catch the exception here and go on so we don't crash.
+            self._handle.send_goal_response(request_header, response_msg)
+        except RCLError:
+            self._logger.warn('Failed to send goal response (the client may have gone away)')
             return
 
-        self._node.get_logger().debug('New goal accepted: {0}'.format(goal_uuid.uuid))
+        if not accepted:
+            self._logger.debug('New goal rejected: {0}'.format(goal_uuid.uuid))
+            return
+
+        self._logger.debug('New goal accepted: {0}'.format(goal_uuid.uuid))
 
         # Provide the user a reference to the goal handle
         await await_or_execute(self._handle_accepted_callback, goal_handle)
 
     async def _execute_goal(self, execute_callback, goal_handle):
         goal_uuid = goal_handle.goal_id.uuid
-        self._node.get_logger().debug('Executing goal with ID {0}'.format(goal_uuid))
+        self._logger.debug('Executing goal with ID {0}'.format(goal_uuid))
 
         try:
             # Execute user callback
@@ -325,28 +334,28 @@ class ActionServer(Waitable):
         except Exception as ex:
             # Create an empty result so that we can still send a response to the client
             execute_result = self._action_type.Result()
-            self._node.get_logger().error('Error raised in execute callback: {0}'.format(ex))
+            self._logger.error('Error raised in execute callback: {0}'.format(ex))
             traceback.print_exc()
 
         # If user did not trigger a terminal state, assume aborted
         if goal_handle.is_active:
-            self._node.get_logger().warning(
+            self._logger.warning(
                 'Goal state not set, assuming aborted. Goal ID: {0}'.format(goal_uuid))
             goal_handle.abort()
 
-        self._node.get_logger().debug(
+        self._logger.debug(
             'Goal with ID {0} finished with state {1}'.format(goal_uuid, goal_handle.status))
 
         # Set result
         result_response = self._action_type.Impl.GetResultService.Response()
         result_response.status = goal_handle.status
         result_response.result = execute_result
-        goal_handle._result_future.set_result(result_response)
+        self._result_futures[bytes(goal_uuid)].set_result(result_response)
 
     async def _execute_cancel_request(self, request_header_and_message):
         request_header, cancel_request = request_header_and_message
 
-        self._node.get_logger().debug('Cancel request received: {0}'.format(cancel_request))
+        self._logger.debug('Cancel request received: {0}'.format(cancel_request))
 
         with self._lock:
             # Get list of goals that are requested to be canceled
@@ -370,7 +379,7 @@ class ActionServer(Waitable):
                     # that will generate an exception from invalid transition.
                     goal_handle._update_state(GoalEvent.CANCEL_GOAL)
                 except RCLError as ex:
-                    self._node.get_logger().debug(
+                    self._logger.debug(
                         'Failed to cancel goal in cancel callback: {0}'.format(ex))
                     # Remove from response since goal has been succeeded
                     cancel_response.goals_canceling.remove(goal_info)
@@ -384,12 +393,12 @@ class ActionServer(Waitable):
         request_header, result_request = request_header_and_message
         goal_uuid = result_request.goal_id.uuid
 
-        self._node.get_logger().debug(
+        self._logger.debug(
             'Result request received for goal with ID: {0}'.format(goal_uuid))
 
         # If no goal with the requested ID exists, then return UNKNOWN status
         if bytes(goal_uuid) not in self._goal_handles:
-            self._node.get_logger().debug(
+            self._logger.debug(
                 'Sending result response for unknown goal ID: {0}'.format(goal_uuid))
             result_response = self._action_type.Impl.GetResultService.Response()
             result_response.status = GoalStatus.STATUS_UNKNOWN
@@ -398,16 +407,24 @@ class ActionServer(Waitable):
 
         # There is an accepted goal matching the goal ID, register a callback to send the
         # response as soon as it's ready
-        self._goal_handles[bytes(goal_uuid)]._result_future.add_done_callback(
+        self._result_futures[bytes(goal_uuid)].add_done_callback(
             functools.partial(self._send_result_response, request_header))
 
     async def _execute_expire_goals(self, expired_goals):
         for goal in expired_goals:
             goal_uuid = bytes(goal.goal_id.uuid)
+            self._goal_handles[goal_uuid].destroy()
             del self._goal_handles[goal_uuid]
+            self.remove_future(self._result_futures[goal_uuid])
+            del self._result_futures[goal_uuid]
 
     def _send_result_response(self, request_header, future):
-        self._handle.send_result_response(request_header, future.result())
+        try:
+            # If the client goes away anytime before this, sending the result response may fail.
+            # Catch the exception here and go on so we don't crash.
+            self._handle.send_result_response(request_header, future.result())
+        except RCLError:
+            self._logger.warn('Failed to send result response (the client may have gone away)')
 
     @property
     def action_type(self):
@@ -493,6 +510,13 @@ class ActionServer(Waitable):
         """Add entities to wait set."""
         with self._lock:
             self._handle.add_to_waitset(wait_set)
+
+    def __enter__(self):
+        return self._handle.__enter__()
+
+    def __exit__(self, t, v, tb):
+        self._handle.__exit__(t, v, tb)
+
     # End Waitable API
 
     def notify_execute(self, goal_handle, execute_callback):
@@ -589,16 +613,12 @@ class ActionServer(Waitable):
 
     def destroy(self):
         """Destroy the underlying action server handle."""
-        if self._handle is None:
-            return
-
         for goal_handle in self._goal_handles.values():
             goal_handle.destroy()
 
+        """Remove the underlying result future."""
+        for result_future in self._result_futures.values():
+            self.remove_future(result_future)
+
         self._handle.destroy_when_not_in_use()
         self._node.remove_waitable(self)
-        self._handle = None
-
-    def __del__(self):
-        """Destroy the underlying action server handle."""
-        self.destroy()
