@@ -12,24 +12,26 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
+#include <pybind11/pybind11.h>
+
+#include <rcl_action/rcl_action.h>
+#include <rcl/error_handling.h>
+#include <rcl/graph.h>
+#include <rcl/types.h>
+#include <rcl_interfaces/msg/parameter_type.h>
+#include <rcl_yaml_param_parser/parser.h>
+#include <rcutils/format_string.h>
+
 #include <limits>
 #include <memory>
+#include <regex>
 #include <stdexcept>
 #include <string>
 #include <utility>
 #include <vector>
 
-#include "pybind11/pybind11.h"
-
-#include "rcl_action/rcl_action.h"
-#include "rcl/error_handling.h"
-#include "rcl/graph.h"
-#include "rcl/rcl.h"
-#include "rcl/types.h"
-#include "rcl_interfaces/msg/parameter_type.h"
-#include "rcl_yaml_param_parser/parser.h"
-#include "rcpputils/scope_exit.hpp"
-#include "rcutils/format_string.h"
+#include <rcpputils/find_and_replace.hpp>
+#include <rcpputils/scope_exit.hpp>
 
 #include "exceptions.hpp"
 #include "logging.hpp"
@@ -260,12 +262,14 @@ _parameter_from_rcl_variant(
  * \param[in] params the parameters for multiple nodes
  * \param[in] pyparameter_cls the Parameter class
  * \param[in] pyparameter_type_cls the Parameter.Type class
+ * \param[in] node_fqn the FQN of node
  * \param[out] pynode_params a dictionary to populate with node names and parameters
  */
 void
 _populate_node_parameters_from_rcl_params(
   const rcl_params_t * params, py::object pyparameter_cls,
-  py::object pyparameter_type_cls, py::dict pynode_params)
+  py::object pyparameter_type_cls, py::dict pynode_params,
+  const char * node_fqn)
 {
   for (size_t i = 0; i < params->num_nodes; ++i) {
     std::string node_name{params->node_names[i]};
@@ -277,6 +281,18 @@ _populate_node_parameters_from_rcl_params(
     if ('/' != node_name.front()) {
       node_name.insert(node_name.begin(), '/');
     }
+
+    if (node_fqn) {
+      // Update the regular expression ["/*" -> "(/\\w+)" and "/**" -> "(/\\w+)*"]
+      std::string regex = rcpputils::find_and_replace(node_name, "/*", "(/\\w+)");
+      if (!std::regex_match(node_fqn, std::regex(regex))) {
+        // No need to parse the items because the user just care about node_fqn
+        continue;
+      }
+
+      node_name = node_fqn;
+    }
+
     auto pynode_name = py::str(node_name);
 
     // make a dictionary for the parameters belonging to this specific node name
@@ -300,12 +316,14 @@ _populate_node_parameters_from_rcl_params(
  * \param[in] args CLI arguments to parse for parameters
  * \param[in] pyparameter_cls the Parameter class
  * \param[in] pyparameter_type_cls the Parameter.Type class
+ * \param[in] node_fqn the FQN of node
  * \param[out] params_by_node_name A Python dict object to place parsed parameters into.
  */
 void
 _parse_param_overrides(
   const rcl_arguments_t * args, py::object pyparameter_cls,
-  py::object pyparameter_type_cls, py::dict pyparams_by_node_name)
+  py::object pyparameter_type_cls, py::dict pyparams_by_node_name,
+  const char * node_fqn)
 {
   rcl_params_t * params = nullptr;
   if (RCL_RET_OK != rcl_arguments_get_param_overrides(args, &params)) {
@@ -314,7 +332,7 @@ _parse_param_overrides(
   if (params) {
     RCPPUTILS_SCOPE_EXIT({rcl_yaml_node_struct_fini(params);});
     _populate_node_parameters_from_rcl_params(
-      params, pyparameter_cls, pyparameter_type_cls, pyparams_by_node_name);
+      params, pyparameter_cls, pyparameter_type_cls, pyparams_by_node_name, node_fqn);
   }
 }
 
@@ -326,35 +344,27 @@ Node::get_parameters(py::object pyparameter_cls)
 
   const rcl_node_options_t * node_options = rcl_node_get_options(rcl_node_.get());
 
-  if (node_options->use_global_arguments) {
-    _parse_param_overrides(
-      &(rcl_node_.get()->context->global_arguments), pyparameter_cls,
-      parameter_type_cls, params_by_node_name);
-  }
-
-  _parse_param_overrides(
-    &(node_options->arguments), pyparameter_cls,
-    parameter_type_cls, params_by_node_name);
-
   const char * node_fqn = rcl_node_get_fully_qualified_name(rcl_node_.get());
   if (!node_fqn) {
     throw RCLError("failed to get node fully qualified name");
   }
+
+  if (node_options->use_global_arguments) {
+    _parse_param_overrides(
+      &(rcl_node_.get()->context->global_arguments), pyparameter_cls,
+      parameter_type_cls, params_by_node_name, node_fqn);
+  }
+
+  _parse_param_overrides(
+    &(node_options->arguments), pyparameter_cls,
+    parameter_type_cls, params_by_node_name, node_fqn);
+
+
   py::str pynode_fqn(node_fqn);
-  py::str pywildcard_name("/**");
   py::dict node_params;
 
-  // Enforce wildcard matching precedence
-  // TODO(cottsay) implement further wildcard matching
-  if (params_by_node_name.contains(pywildcard_name)) {
-    node_params = params_by_node_name[pywildcard_name];
-  }
   if (params_by_node_name.contains(pynode_fqn)) {
-    // TODO(sloretz) py::dict should expose dict.update()
-    py::dict node_specific_params = params_by_node_name[pynode_fqn];
-    for (const std::pair<py::handle, py::handle> & key_value : node_specific_params) {
-      node_params[key_value.first] = key_value.second;
-    }
+    node_params = params_by_node_name[pynode_fqn];
   }
 
   return node_params;
