@@ -13,11 +13,14 @@
 # limitations under the License.
 
 import asyncio
+import os
 import threading
 import time
 import unittest
+import warnings
 
 import rclpy
+from rclpy.executors import Executor
 from rclpy.executors import MultiThreadedExecutor
 from rclpy.executors import ShutdownException
 from rclpy.executors import SingleThreadedExecutor
@@ -34,6 +37,7 @@ class TestExecutor(unittest.TestCase):
     def tearDown(self):
         self.node.destroy_node()
         rclpy.shutdown(context=self.context)
+        self.context.destroy()
 
     def func_execution(self, executor):
         got_callback = False
@@ -128,6 +132,32 @@ class TestExecutor(unittest.TestCase):
 
         assert not got_callback
 
+    def test_multi_threaded_executor_num_threads(self):
+        self.assertIsNotNone(self.node.handle)
+
+        # check default behavior, either platform configuration or defaults to 2
+        executor = MultiThreadedExecutor(context=self.context)
+        if hasattr(os, 'sched_getaffinity'):
+            platform_threads = len(os.sched_getaffinity(0))
+        else:
+            platform_threads = os.cpu_count()
+        self.assertEqual(platform_threads, executor._executor._max_workers)
+        executor.shutdown()
+
+        # check specified thread number w/o warning
+        executor = MultiThreadedExecutor(num_threads=3, context=self.context)
+        self.assertEqual(3, executor._executor._max_workers)
+        executor.shutdown()
+
+        # check specified thread number = 1, expecting UserWarning
+        with warnings.catch_warnings(record=True) as w:
+            warnings.simplefilter('always', category=UserWarning)
+            executor = MultiThreadedExecutor(num_threads=1, context=self.context)
+            self.assertEqual(1, executor._executor._max_workers)
+            executor.shutdown()
+            assert len(w) == 1
+            assert issubclass(w[0].category, UserWarning)
+
     def test_multi_threaded_executor_executes(self):
         self.assertIsNotNone(self.node.handle)
         executor = MultiThreadedExecutor(context=self.context)
@@ -456,6 +486,97 @@ class TestExecutor(unittest.TestCase):
             executor.shutdown()
             self.node.destroy_timer(tmr)
 
+    def shutdown_executor_from_callback(self):
+        """https://github.com/ros2/rclpy/issues/944: allow for executor shutdown from callback."""
+        self.assertIsNotNone(self.node.handle)
+        timer_period = 0.1
+        executor = SingleThreadedExecutor(context=self.context)
+        shutdown_event = threading.Event()
+
+        def timer_callback():
+            nonlocal shutdown_event, executor
+            executor.shutdown()
+            shutdown_event.set()
+
+        tmr = self.node.create_timer(timer_period, timer_callback)
+        executor.add_node(self.node)
+        t = threading.Thread(target=executor.spin, daemon=True)
+        t.start()
+        self.assertTrue(shutdown_event.wait(120))
+        self.node.destroy_timer(tmr)
+
+    def test_context_manager(self):
+        self.assertIsNotNone(self.node.handle)
+
+        executor: Executor = SingleThreadedExecutor(context=self.context)
+
+        with executor as the_executor:
+            # Make sure the correct instance is returned
+            assert the_executor is executor
+
+            assert not executor._is_shutdown, 'the executor should not be shut down'
+
+        assert executor._is_shutdown, 'the executor should now be shut down'
+
+        # Make sure it does not raise (smoke test)
+        executor.shutdown()
+
+    def test_single_threaded_spin_once_until_future(self):
+        self.assertIsNotNone(self.node.handle)
+        executor = SingleThreadedExecutor(context=self.context)
+
+        future = Future(executor=executor)
+
+        # Setup a thread to spin_once_until_future_complete, which will spin
+        # for a maximum of 10 seconds.
+        start = time.time()
+        thread = threading.Thread(target=executor.spin_once_until_future_complete,
+                                  args=(future, 10))
+        thread.start()
+
+        # Mark the future as complete immediately
+        future.set_result(True)
+
+        thread.join()
+        end = time.time()
+
+        time_spent = end - start
+
+        # Since we marked the future as complete immediately, the amount of
+        # time we spent should be *substantially* less than the 10 second
+        # timeout we set on the spin.
+        assert time_spent < 10
+
+        executor.shutdown()
+
+    def test_multi_threaded_spin_once_until_future(self):
+        self.assertIsNotNone(self.node.handle)
+        executor = MultiThreadedExecutor(context=self.context)
+
+        future = Future(executor=executor)
+
+        # Setup a thread to spin_once_until_future_complete, which will spin
+        # for a maximum of 10 seconds.
+        start = time.time()
+        thread = threading.Thread(target=executor.spin_once_until_future_complete,
+                                  args=(future, 10))
+        thread.start()
+
+        # Mark the future as complete immediately
+        future.set_result(True)
+
+        thread.join()
+        end = time.time()
+
+        time_spent = end - start
+
+        # Since we marked the future as complete immediately, the amount of
+        # time we spent should be *substantially* less than the 10 second
+        # timeout we set on the spin.
+        assert time_spent < 10
+
+        executor.shutdown()
+
 
 if __name__ == '__main__':
     unittest.main()
